/*
 * sshread.c
 *
 * routines of input, and formatting according to the styles
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97 Akim Demaille, Miguel Santana
 * $Id: sshread.c,v 1.46 1997/09/30 15:45:57 demaille Exp $
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * $Id: sshread.c,v 1.46 1997/09/30 15:45:57 demaille Exp $
 */

#include "a2ps.h"
#include "sshread.h"
#include "ssheet.h"
#include "routines.h"
#include "buffer.h"
#include "jobs.h"
#include "psgen.h"

/*
 * Use the information offered by main.c
 */
extern struct a2ps_job * job;
extern int strip_level;

/*
 * Structure in which is stored the result of a parsing
 */
struct token {
  struct re_registers * registers;
  struct darray * destinations;
};

#define token_dest(_i_)	\
    ((struct faced_string *) token->destinations->content[_i_])

static inline struct token *
token_new (void)
{
  struct token * res = ALLOC (struct token, 1);
  res->registers = ALLOC (struct re_registers, 1);
  res->registers->start = ALLOC (regoff_t, 30);
  res->registers->end = ALLOC (regoff_t, 30);
  return res;
}

static inline struct darray *
destination_plain_new (void)
{
  return destinations_new_single (NULL, 0, Plain);
}

static inline void
token_free (struct token * token)
{
  FREE (token);
}

/* Where the token and its attributes are stored */
static struct token * token = NULL;
#define token_set_registers(_start_, _len_)		\
 do {							\
  token->registers->start [0] = _start_;		\
  token->registers->end [0] = _start_ + _len_;		\
 } while (0)

#define token_start(_i_)		\
  token->registers->start [token_dest(_i_)->reg_ref]

#define token_end(_i_)		\
  token->registers->end [token_dest(_i_)->reg_ref]

static struct darray * plain_destination = NULL;

/****************************************************************/
/*		pretty printing service routines	       	*/
/****************************************************************/
/*
 * Set correctly the reg registers
 */
static inline void
match_word (struct buffer * buffer, struct style_sheet * sheet)
{
  int i=1;
  
  while (sheet->alpha2[*(buffer->content + buffer->curr + i)])
    i++;
  
  token->destinations = plain_destination ;
  token_dest (0) -> face = Plain;
  token->registers->start [0] = buffer->curr;
  token->registers->end [0] = buffer->curr + i;

  buffer->curr += i;
}

/****************************************************************/
/*			lexical analysis routines	       	*/
/****************************************************************/
#define word_regexp(_i_)	\
   (((struct symbol *) words->regexps->content[_i_]))
/* We give the whole sheet, because alphabets are needed */
static inline int
match_keyword (struct buffer * buffer, 
	       struct words * words,
	       uchar * alphabet)
{
  struct symbol ** key;
  uchar * string = buffer->value + buffer->curr;
  int i, res;

  /* First try the regexps */
  for (i = 0 ; i < words->regexps->len ; i++) {
    switch (res = (re_match (word_regexp(i)->regex,
			     (char *) buffer->value, buffer->len, buffer->curr,
			     token->registers))) {
    case -2: 
      fprintf (stderr, "An error occured while matching\n");
      break;
    case -1:
      continue;
    default:
      token->destinations = word_regexp(i)->destination;
      buffer->curr += res;
      return 1;
    }
  }
  
  /* Then try the words */
  if (!words->min [*string])
    return 0;
  for (key = words->max [*string]
	 ; words->min [*string] <= key
	 ; key --)
    {
      if (ustrprefix ((*key)->word, string)
	  && !alphabet[string[ustrlen ((*key)->word)]]) {
	token->destinations = (*key)->destination;
	token_set_registers (buffer->curr, ustrlen ((*key)->word));
	buffer->curr += ustrlen ((*key)->word);
	return 1;
      }
    }
  return 0;
}

static inline int
match_operator (struct buffer * buffer, 
		struct words * words)
{
  struct symbol ** key;
  uchar * string = buffer->value + buffer->curr;
  int i, res;

  /* First try the regexps */
  for (i = 0 ; i < words->regexps->len ; i++) {
    switch (res = (re_match (word_regexp(i)->regex,
			     (char *) buffer->value, buffer->len, buffer->curr,
			     token->registers))) {
    case -2: 
      fprintf (stderr, "An error occured while matching\n");
      break;
    case -1:
      continue;
    default:
      token->destinations = word_regexp(i)->destination;
      buffer->curr += res;
      return 1;
    }
  }
  
  /* Then try the words */
  if (!words->min [*string])
    return 0;
  for (key = words->max [*string]
	 ; words->min [*string] <= key
	 ; key --) {
    if (ustrprefix((*key)->word, string)) {
      token->destinations = (*key)->destination;
      token_set_registers (buffer->curr, ustrlen ((*key)->word));
      buffer->curr += ustrlen ((*key)->word);
      return 1;
    }
  }
  
  return 0;
}

/* 
 * If buffer+*curr begins with a sequence, return that sequence.
 * Otherwise NULL
 */
#ifdef SEQ
#  undef SEQ
#endif
#define SEQ(_i_) 	\
   ((struct sequence *) sheet->sequences->content [i])
static inline struct sequence *
match_sequence (struct buffer * buffer, struct style_sheet * sheet)
{
  int i;
  int res;
  uchar * string = buffer->value + buffer->curr;
  
  for (i = 0 ; i < sheet->sequences->len ; i++)
    {
      if (SEQ(i)->open->regex)
	{
	  /* The regexp patterns */
	  res = re_match (SEQ(i)->open->regex,
			  (char *) buffer->value, buffer->len, buffer->curr,
			  token->registers);
	  switch (res) {
	  case -2: 
	    fprintf (stderr, "An error occured while matching\n");
	    break;
	  case -1:
	    continue;
	  default:
	    token->destinations = SEQ(i)->open->destination;
	    buffer->curr += res;
	    return SEQ(i);
	  }
	}
      else 
	{
	  /* It's a string */
	  if (ustrprefix (SEQ(i)->open->word, string)) {
	    token_set_registers (buffer->curr, ustrlen (SEQ(i)->open->word));
	    token->destinations = SEQ(i)->open->destination;
	    buffer->curr += ustrlen (SEQ(i)->open->word);
	    return SEQ(i);
	  }
	}
    }
  return NULL;
}

/* 
 * return TRUE if buffer+*curr begins with a symbol contained in array
 */
#define symbol_num(i) ((struct symbol *)array->content[i])
static inline int
match_symbols_array (struct buffer * buffer, struct darray * array)
{
  int i, res;
  uchar * string = buffer->value + buffer->curr;
  
  for (i = 0 ; i < array->len ; i++)
    {
      if (symbol_num (i)->regex)
	{
	  /* The regexp patterns */
	  res = re_match (symbol_num (i)->regex,
			  (char *) buffer->value, buffer->len, buffer->curr,
			  token->registers);
	  switch (res) {
	  case -2: 
	    fprintf (stderr, "An error occured while matching\n");
	    break;
	  case -1:
	    continue;
	  default:
	    token->destinations = symbol_num(i)->destination;
	    buffer->curr += res;
	    return 1;
	  }
	}
      else 
	{
	  /* It's a strings */
	  if (ustrprefix (symbol_num (i)->word, string)) {
	    token->destinations = symbol_num (i)->destination;
	    token_set_registers (buffer->curr, ustrlen (symbol_num (i)->word));
	    buffer->curr += ustrlen (symbol_num (i)->word);
	    return 1;
	  }
	}
    }
  return 0;
}

/*
 * Put in token the token recognized.
 * The number of token read, 0 if nothing left
 */
static inline int 
ssh_get_token (struct buffer * buffer, struct style_sheet * sheet)
{
  static int return_to_plain = FALSE;
  /* NULL if not in a sequence currently */
  static struct sequence * sequence = NULL;

  if (buffer_is_empty (buffer)) 
    {
      buffer_get (buffer);

      /* We don't trust liba2ps for the line numbers, because
       * if a2ps skips some lines (e.g., --strip-level, or INVISIBLE),
       * liba2ps will number upon output lines, not imput lines, 
       * which is what is expected */
      (CURRENT_FILE (job))->lines = buffer->line;
      
      if (buffer->len == 0) {
	/* end of file: reset values */
	/* If this is a new file, it must not depend on the trailling
	 * parameters of the previous file */
	sequence = NULL;
	return_to_plain = FALSE;
	return 0;
      }
    }

  if (return_to_plain) {
    return_to_plain = FALSE;
    token->destinations = plain_destination;
    token_dest (0) -> face = Plain;
  }

  if (sequence) {
    /* escape: not converted when in a sequence */
    if (match_symbols_array (buffer, sequence->exceptions))
      return 1;
    /* end of sequence ? */
    if (match_symbols_array (buffer, sequence->close))
      {
	return_to_plain = TRUE;
        sequence = NULL;
	return 1;
      }
    /* We are in a sequence not to be closed yet.
     * Advance of 1 char */
    token->destinations = plain_destination;
    token_dest (0) -> face = sequence->face;
    token->registers->start [0] = buffer->curr++;
    token->registers->end [0] = buffer->curr;
    return 1;
  } else { /* (not in sequence) */
    if ((sequence = match_sequence (buffer, sheet)))
      return 1;
    else if (sheet->alpha1[buffer->content[buffer->curr]]) {  
      /* we are in a word since this was a char belonging to the
       * first alphabet */
      if (match_keyword (buffer, sheet->keywords, sheet->alpha2)
	  || match_operator (buffer, sheet->operators)) {
	return_to_plain = TRUE;
	return 1;
      } else {
	/* since some characters may be used inside an identifier
	 * (eg, x' = x in claire) but can also be used to open
	 * a special sequence (eg, 'x' in claire), then we must read
	 * the whole word, and print in.
	 */
	match_word (buffer, sheet);
	return 1;
      }
    } else if (match_operator (buffer, sheet->operators)) {
      return_to_plain = TRUE;
      return 1;
    } else
      /* Why did I put this before ?
       * token->face = Plain */ ;
  }
  
  /* We did not recognize something special */
  token->destinations = plain_destination;
  token->registers->start [0] = buffer->curr++;
  token->registers->end [0] = buffer->curr;
  return 1;
}

#define GRAB_TAG(_tag_)					\
  do {							\
    ustrncat (_tag_,					\
	      buffer->content + token_start (i),	\
	      token_end (i) - token_start (i));		\
  } while (0)

/*
 * ssh-Pretty print a file to postscript
 */
void
ssh_print_postscript (struct a2ps_job * Job, 
		      struct buffer * buffer,
		      struct style_sheet * sheet)
{
  face_t face = Plain;
  /* To grab the encoding switching instruction */
  uchar bufenc[512];
  int grabbing_encoding = FALSE;
  int i;

  *bufenc = NUL;

  /* I must do this because of the broken handling of the registers in
   * regex.  Until I find a way to ensure enough place in the
   * registers, without having to trust regex. */
  if (!token) {
    token = token_new ();
    plain_destination = destination_plain_new ();
  }

  while (ssh_get_token (buffer, sheet) != 0)
    for (i = 0 ; i < token->destinations->len ; i++)
      {
	/* Is a new face ? */
	if (face != token_dest (i)->face) {
	  /* Reset dynamic markers */
	  if (token_dest (i)->face & TAG1)
	    *Job->tag1 = NUL;
	  if (token_dest (i)->face & TAG2)
	    *Job->tag2 = NUL;
	  if (token_dest (i)->face & TAG3)
	    *Job->tag3 = NUL;
	  if (token_dest (i)->face & TAG4)
	    *Job->tag4 = NUL;
	  if (grabbing_encoding && (!(token_dest (i)->face & ENCODING_TAG))) {
	    /* Grabbing of the encoding name is completed */
	    struct encoding_entry * newenc;
	    newenc = get_encoding_by_alias (job, (char *) bufenc);
/*	    encoding_build_faces_wx (job, newenc);*/
	    if (!newenc)
	      error (0, 0, _("unknown encoding: `%s', ignored"), bufenc);
	    else
	      ps_switch_encoding (Job, newenc);
	    * bufenc = NUL;
	    grabbing_encoding = FALSE;
	  }
	  face = token_dest (i)->face;
	}

	/* See if there are some information to grab */
	if (token_dest (i)->string)
	  {
	    if (face & TAG1)
	      ustrcat (Job->tag1, token_dest (i)->string);
	    if (face & TAG2)
	      ustrcat (Job->tag2, token_dest (i)->string);
	    if (face & TAG3)
	      ustrcat (Job->tag3, token_dest (i)->string);
	    if (face & TAG4)
	      ustrcat (Job->tag4, token_dest (i)->string);
	    /* Grab the dynamic encodings */
	    if (face & ENCODING_TAG) {
	      grabbing_encoding = TRUE;
	      ustrcat (bufenc, token_dest (i)->string);
	    }
	  } else {
	    if (face & TAG1)
	      GRAB_TAG (Job->tag1);
	    if (face & TAG2)
	      GRAB_TAG (Job->tag2);
	    if (face & TAG3)
	      GRAB_TAG (Job->tag3);
	    if (face & TAG4)
	      GRAB_TAG (Job->tag4);
	    /* Grab the dynamic encodings */
	    if (face & ENCODING_TAG) {
	      grabbing_encoding = TRUE;
	      GRAB_TAG (bufenc);
	    }
	  }	    
	/* if the face is INVISIBLE, forget everything */
	/* if strip, skip (strong_)comments */
	/* FIXME: A better scheme, with just masks should be possible */
	if ((face & INVISIBLE)
	    || ((strip_level == 1 || strip_level == 3)
		&& face == Comment)
	    || ((strip_level == 2 || strip_level == 3)
		&& face == Comment_strong))
	  continue;
	
	/* Return what's new */ 
	if (token_dest (i)->string) {
	  ps_print_string (Job, token_dest (i)->string, 
			   face & FACE_MASK);
	} else {
	  ps_print_buffer (Job, buffer->content, 
			   token_start (i), token_end (i),
			   face & FACE_MASK);
	}
      }
  
  /* FIXME: Find a way to remove the token some day
  token_free (token);
*/
}
